from os import listdir
from os.path import isfile, join
from AutoRPE.UtilsRPE.Functions import calls_to_routines, calls_to_functions, is_keyword_argument, is_a_number, \
    find_called_routine, find_call_arguments, find_called_function, \
    type_of_contents, in_which_block, clean_contents, \
    match_argument_types_and_interface, load_vault, load_variable_list
from AutoRPE.UtilsRPE.Classes import ImplementRPE, Interface
import warnings


def code_check_file(file, vault):
    try:
        current_block = vault[file.split("/")[-1].replace(".f90", "")].main()
    except AttributeError:
        warnings.warn("No module %s" % file)
        return
    f = ImplementRPE(file, None)

    f.pre_process()
    # Loop through lines
    for line_n, line in enumerate(f.lines):
        # Check if there are routine/function calls
        called_procedures = []
        procedure_calls = []
        routine_calls = calls_to_routines(line)
        if routine_calls:
            for routine_call in routine_calls:
                called_routine = find_called_routine(routine_call)
                if called_routine == "date_and_time":
                    continue
                procedure_calls.append(routine_call)
                called_procedures.append(called_routine)
        else:
            function_calls = calls_to_functions(line, vault)
            if function_calls:
                for function_call in function_calls:
                    procedure_calls.append(function_call)
                    called_procedures.append(find_called_function(function_call))

        if called_procedures:
            for procedure_call, called_procedure in zip(procedure_calls, called_procedures):
                called_arguments = find_call_arguments(procedure_call)
                if not called_arguments:
                    continue

                procedure = vault.get_procedure(called_procedure)
                indices_to_ignore = []

                if isinstance(procedure, Interface):
                    argument_types = [type_of_contents(argument, routine=current_block, vault=vault) for argument in
                                      called_arguments]
                    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in
                                         called_arguments]
                    # Checking if some of the arguments does not necessarily need to be bounded to the dummy
                    # because there's already an existing version within the same interface with a different precision
                    for _in, _type in enumerate(argument_types):
                        new_argument_types = argument_types[:]
                        if _type == "single":
                            new_argument_types[_in] = "double"
                        elif _type == "double":
                            new_argument_types[_in] = "single"
                        else:
                            continue
                        try:
                            indices_to_ignore.append(_in)
                        except AssertionError:
                            continue
                    # Classical match
                    try:
                        procedure = match_argument_types_and_interface(argument_types, procedure,
                                                                       keyword_arguments=keyword_arguments,
                                                                       current_block=current_block, vault=vault)
                    except AssertionError:
                        print("INTERFACE: ", current_block.name, procedure.name, [at for at in argument_types],
                              "file: ", file.split("/")[-1])

                if procedure is None:
                    continue
                for index, argument in enumerate(called_arguments):
                    if index in indices_to_ignore:
                        continue

                    if is_keyword_argument(argument):
                        [key, value] = argument.split("=", 1)
                        key = key.strip()
                        value = value.strip()
                        var_type = type_of_contents(value, routine=current_block, vault=vault)
                    else:
                        var_type = type_of_contents(argument, routine=current_block, vault=vault)

                    if var_type not in ["single", "double", "rpe_var"] or is_a_number(argument):
                        continue

                    # in case is Key, using index will retrieve the wrong variable
                    if is_keyword_argument(argument):
                        dummy_var = procedure.get_variable_by_name(key)
                    else:
                        dummy_var = procedure.get_variable_by_position(index)

                    if var_type == dummy_var.type:
                        continue
                    else:
                        var_name = clean_contents(argument)
                        #if not dummy_var.intent == "in":
                        print("ARGUMENT => ", var_name, " is ", var_type, "instead of  ", dummy_var.type,
                                  "in ", called_procedure, "file: ", file.split("/")[-1])

        # If not, look if the code is entering or exiting a routine block
        current_block = in_which_block(line, current_block, vault=vault)


if __name__ == "__main__":
    # Getting Command line options
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-v", "--vault", dest="vault", default=None, metavar="FILE")
    parser.add_option("-l", "--list", dest="variable_list_filename", default=None, metavar="FILE")
    parser.add_option("-s", "--input-source", dest="input_source", default=None, metavar="FILE")
    parser.add_option("-w", "--working-precision", dest="working_precision", default="single")
    parser.add_option("-t", "--target-precision", dest="target_precision", default="double")

    (options, args) = parser.parse_args()
    vault_path = options.vault
    wp = options.working_precision
    target_precision = options.target_precision
    path_to_input_sources = options.input_source
    variable_list_filename = options.variable_list_filename

    if path_to_input_sources is None or vault_path is None or variable_list_filename is None:
        print("\nOne of the arguments -s/--input-source, -v/--vault is mandatory and -l/--list are mandatory\n")
        parser.print_help()
        exit(1)

    print("Checking coherency between list in %s and code in:\n   %s\n\n" % (
        variable_list_filename, path_to_input_sources))

    # Loading vault
    vault = load_vault(vault_path)

    # Getting variables
    variable_ids = load_variable_list(variable_list_filename)

    # Modify the type of the variables selected through the filename
    variables = [vault.get_variable_by_id(_id) for _id in variable_ids]
    for var in variables:
        var.type = target_precision

        # Find files in sources
    all_sources = [f for f in listdir(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90")]
    # Process Sources!
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        code_check_file(join(path_to_input_sources, filename), vault)
